package team.app.system;

import java.util.ArrayList;
import java.util.List;

import org.jscience.mathematics.numbers.Number;
import org.jscience.mathematics.vectors.Vector;

import team.app.App;
import team.app.IConfig;
import team.app.system.loader.IScehduleLoader;
import team.app.system.loader.ScheduleData;
import team.app.system.loader.TableScheduleLoader;
import team.app.wp.WPUtils;
import team.app.wp.WayPoints;
import team.gui.Coord;
import team.sim.models.Time;

public class ScheduleCalc
{
	private IConfig app;
	
	private WayPoints wps;
	private List data = new ArrayList();
	
	private final IScehduleLoader dloader;

	private double total_length;
	private double sch_start_time;
	private double sch_end_time;
	
	private double obs_time;
	private double obs_length;
	private double obs_adherence;
	private Coord obs_pos = new Coord(0,0);

	private String routeID;	
	private int tripID;
	

	public ScheduleCalc(IConfig app, IScehduleLoader dloader)
	{
		this.app = app;
		this.dloader = dloader;
	}
	
	public void loadRoute(String rid, int tripID, double lsch_start_time)
	{
		routeID = rid;
		this.tripID = tripID;
		
		if(rid == null)
		{
			wps.clear();
			total_length = 0;
			data.clear();
			sch_start_time = Double.MAX_VALUE;
			obs_adherence = 0;
			return;
		}
		
		wps = WPUtils.loadRoute(rid, app);		
		total_length = WPUtils.getTotalLength(wps); //WPUtils.getRouteLengthExEnds(rid, app);		
		data = dloader.loadRouteData(rid, tripID, lsch_start_time);
		
		
		sch_start_time = lsch_start_time;
		sch_end_time = ((ScheduleData)data.get(data.size()-1)).sch_time + sch_start_time;
		
	}
	
	public void regObservation(Coord pos, double time)
	{
		obs_pos.set(pos);
		obs_time = time;
		
		if(time < sch_start_time)
		{
			obs_length = 0;
			obs_adherence = 0;
		}
		else
		{
			double est_length = estimateLength(time);
			obs_length = WPUtils.getDistance(wps, pos, est_length);
			obs_adherence = getScheduleTime(obs_length) - time;
			
			//System.out.println(est_length - obs_length);
		}
	}
	
	public double getNormalComp(double time, Coord obs_pos)
	{
		Vector va = estimatePosition(time + obs_adherence).toVector();
		Vector vb = estimatePosition(time + obs_adherence+1).toVector();
		
		Vector vd = vb.minus(va);
		Vector vp = obs_pos.toVector().minus(va);
		
		double est_length_comp = ((Number)vp.times(vd)).doubleValue();
		est_length_comp /= Coord.length(vd);
		
		if(Double.isNaN(est_length_comp))
			est_length_comp = 0;
		
		return est_length_comp;
	}
	
	public double getScheduleTime(double length)
	{
		if(data.isEmpty())
			return sch_start_time;
		
		ScheduleData a = (ScheduleData) data.get(data.size()-1);
		if(length >= a.acu_length)
			return a.sch_time + sch_start_time;
		
		ScheduleData b = (ScheduleData) data.get(0);
		if(length <= b.acu_length)
			return b.sch_time + sch_start_time;
		
		int findex = -1;
		for(int i = 0; i < data.size(); i++)
		{
			a = (ScheduleData) data.get(i);
			if(length <= a.acu_length)
			{
				findex = i-1;
				break;
			}
		}
		
		a = (ScheduleData) data.get(findex);
		b = (ScheduleData) data.get(findex+1);
		
		return sch_start_time + interpolate(a.sch_time, b.sch_time, ratio(a.acu_length,b.acu_length,length));
	}
	
	public double getScheduleLength(double time)
	{		
		if(data.isEmpty())
			return 0;
		
		ScheduleData a = (ScheduleData) data.get(data.size()-1);
		if(time >= a.sch_time + sch_start_time)
			return a.acu_length;
		
		ScheduleData b = (ScheduleData) data.get(0);
		if(time <= b.sch_time + sch_start_time)
			return b.acu_length;
		
		int findex = -1;
		for(int i = 1; i < data.size(); i++)
		{
			a = (ScheduleData) data.get(i);
			if(time <= a.sch_time + sch_start_time)
			{
				findex = i-1;
				break;
			}
		}
		
		if(findex == -1)
		{
			System.out.println("sch_time: "+(time - sch_start_time));
			System.out.println(""+(ScheduleData) data.get(data.size()-1));
			System.out.println(""+(ScheduleData) data.get(0));
		}
		
		a = (ScheduleData) data.get(findex);
		b = (ScheduleData) data.get(findex+1);
		
		return interpolate(a.acu_length, b.acu_length, ratio(a.sch_time, b.sch_time, time - sch_start_time));
	}
	
	public Coord getScheduledPosition(double time)
	{
		if(wps.isEmpty())
			return obs_pos;
		
		double length = getScheduleLength(time);
		return WPUtils.getPosition(wps, length);
	}
	
	/*public double getAdherence(Coord pos, double time)
	{
		if(time < sch_start_time)
			return 0;
		
		double length = WPDriver.getDistance(wps, pos, estimateLength(time));
		return (getScheduleTime(length)-time);
	}
	
	public Coord getCoordAdherence(double adherence, double time)
	{
		return getScheduledPosition(time + adherence);
	}*/
	
	public double getObsAdherence()
	{
		return obs_adherence;
	}
	
	public double getRouteTime(double time)
	{
		if(time < sch_start_time)
			return 0;
		else
			return time - sch_start_time;
	}
	
	public int getObsDay()
	{
		return (int)(obs_time / (24*60*60));
	}
	
	public String getRouteID()
	{
		return routeID;
	}
	
	public int getTripID()
	{
		return tripID;
	}
	
	public double getScheduleTime()
	{
		return sch_end_time - sch_start_time;
	}
	
	public double getScheduleEndTime()
	{
		return sch_end_time;
	}
	
	public double getScheduleStartTime()
	{
		return sch_start_time;
	}
	
	public double getRouteLength()
	{
		return total_length;
	}
	
	public double getObsRouteTime()
	{
		return getRouteTime(getObsTime());
	}
	
	public double getObsLengthComplete()
	{
		return estimateLength(getObsTime());
	}
	
	public double getObsDelay()
	{
		return obs_adherence * -1;
	}
	
	public Coord getObsPosition()
	{
		return obs_pos;
	}
	
	public double getObsTime()
	{
		return obs_time;
	}
	
	public double getObsNLengthComplete()
	{
		return estimateLength(getObsTime()) / getRouteLength();
	}
	
	public Coord estimatePosition(double time)
	{
		return getScheduledPosition(time-estimateDelay(time));
	}
	
	public double estimateLength(double time)
	{
		return getScheduleLength(time-estimateDelay(time));
	}
	
	public double estimateLengthLeft(double time)
	{
		return total_length - estimateLength(time);
	}
	
	public double estimateAdherence(double time)
	{
		return getObsAdherence();
	}
	
	public double estimateDelay(double time)
	{
		//should depend om time -> lastObsedDelay - time_since_obs * factor
		return getObsDelay();
	}
	
	public static double ratio(double a, double b, double value)
	{
		if(value <= a)
			return 0;
		else if(value >= b)
			return 1;
		
		return (value-a)/(b-a);
	}
	
	public static double interpolate(double a, double b, double ratio)
	{
		return a+ratio*(b-a);
	}
	
	public static Coord interpolate(Coord a, Coord b, double ratio)
	{
		Coord tmp = new Coord(0,0);
		tmp.x = interpolate(a.x, b.x, ratio);
		tmp.y = interpolate(a.y, b.y, ratio);
		return tmp;
	}
	
	public static void main(String[] args) throws Exception
	{
		App app = new App();
		ScheduleCalc p = new ScheduleCalc(app, new TableScheduleLoader(app));
		p.loadRoute("52.P",29,Time.toDouble(12, 20, 0));
		
		double test = Time.toDouble(12, 29, 30);
		
		Coord pos = p.getScheduledPosition(test);
		double length = p.getScheduleLength(test);
		double time = p.getScheduleTime(length);
	
		
		java.lang.System.out.println("pos: "+pos);
		java.lang.System.out.println("length: "+length);
		java.lang.System.out.println("sch: " +Time.format(time));
		
		p.regObservation(p.getScheduledPosition(test-30), time);
		System.out.println("delay: " + p.getObsDelay());
		System.out.println("est p: " + p.estimatePosition(test));
		System.out.println("est l: " + p.estimateLength(test));

		
		//System.out.println(WPDriver.getDistance(p.driver.getWayPoints(), pos, 1000));
	}
	
	
}
